%% Section 10.2.1
% Load data
load ionosphere;
%%
% For reproducibility, set the random number generator seed for data
% partitioning.
rng(1);
%%
% Build a cross-validated pseudo linear discriminant
cvplda = ClassificationDiscriminant.fit(X,Y,'DiscrimType','pseudoLinear','crossval','on');
%%
% Grow a cross-validated ensemble of decision stumps boosted by AdaBoostM1
cvada = fitensemble(X,Y,'AdaBoostM1',200,'Tree','crossval','on');
%%
% Computed predicted scores for out-of-fold data
[~,SfitLDA] = kfoldPredict(cvplda);
[~,SfitAda] = kfoldPredict(cvada);
%%
% Obtain FPR and TPR choosing 'good returns' as the positive class
[fprLDA,tprLDA] = perfcurve(Y,SfitLDA(:,2),'g');
[fprAda,tprAda] = perfcurve(Y,SfitAda(:,2),'g');
%%
% Plot the ROC curves
figure;
plot(fprLDA,tprLDA,'b--');
hold;
plot(fprAda,tprAda,'r-.');
line([0 1],[0 1],'color','c');
xlabel('False positive rate');
ylabel('True positive rate');
legend('Pseudo LDA','Boosted stumps','Fair coin toss','Location','SE');

%% Section 10.2.5

%% Estimate ROC by kernel smoothing
%%
% Find good and bad radar returns
isGood = strcmp(Y,'g');
isBad = strcmp(Y,'b');
%%
% Count numbers of good and bad radar returns
Ngood = sum(isGood);
Nbad = sum(isBad);
%%
% Obtain quantiles for kernel bandwidth estimation
Qgood = quantile(SfitLDA(isGood,2),[0.25 0.75]);
Qbad = quantile(SfitLDA(isBad,2),[0.25 0.75]);
%%
% Estimate optimal kernel bandwidth
Hgood = 1.58*(Qgood(2)-Qgood(1))/1.34/Ngood^(1/3)
Hbad = 1.58*(Qbad(2)-Qbad(1))/1.34/Nbad^(1/3)

thre = 0:0.001:1;
Fnaive = ksdensity(SfitLDA(isBad,2),thre,'width',Hbad,'function','cdf');
F = ksdensity(SfitLDA(isBad,2),thre,'width',0.02,'function','cdf');
Gnaive = ksdensity(SfitLDA(isGood,2),thre,'width',Hgood,'function','cdf');
G = ksdensity(SfitLDA(isGood,2),thre,'width',0.02,'function','cdf');

figure;
plot(fprLDA,tprLDA);
hold;
plot(1-Fnaive,1-Gnaive,'r--');
plot(1-F,1-G,'c-.');
legend('Empirical','Naive bandwidth','Optimal bandwidth','Location','SE');
xlabel('False positive rate');
ylabel('True positive rate');

%% Kernel smoothing after a score transform
%%
% Recompute classification scores after applying inverse logit
% transformation. Select the 2nd column of SfitLDA which corresponds to
% good radar returns ('g' class).
cvplda.ScoreTransform = 'invlogit';
[~,SfitLDA] = kfoldPredict(cvplda);
SfitLDA = SfitLDA(:,2);
%%
% Compute empirical curve with score thresholds
[fprLDA,tprLDA,threLDA] = perfcurve(Y,SfitLDA,'g');
%%
% Obtain quantiles for kernel bandwidth estimation
Qgood = quantile(SfitLDA(isGood),[0.25 0.75]);
Qbad = quantile(SfitLDA(isBad),[0.25 0.75]);
%%
% Estimate optimal kernel bandwidth
Hgood = 1.58*(Qgood(2)-Qgood(1))/1.34/Ngood^(1/3)
Hbad = 1.58*(Qbad(2)-Qbad(1))/1.34/Nbad^(1/3)
%%
% Compute smoothed TNR and FNR at 1000 points uniformly spaced from minimal
% to maximal score
thre = linspace(min(SfitLDA),max(SfitLDA),1000);
F = ksdensity(SfitLDA(isBad),thre,'width',Hbad,'function','cdf');   % TNR
G = ksdensity(SfitLDA(isGood),thre,'width',Hgood,'function','cdf'); % FNR
%%
% Plot empirical and smoothed FPR
figure;
plot(threLDA,fprLDA);
hold;
plot(thre,1-F,'r--');
legend('Empirical','Smoothed','Location','SW');
xlabel('Threshold');
ylabel('False positive rate');
%%
% Plot empirical and smoothed TPR
figure;
plot(threLDA,tprLDA);
hold;
plot(thre,1-G,'r--');
legend('Empirical','Smoothed','Location','SW');
xlabel('Threshold');
ylabel('True positive rate');
%%
% Plot empirical and smoothed ROC curve
figure;
plot(fprLDA,tprLDA);
hold;
plot(1-F,1-G,'r--');
legend('Empirical','Smoothed','Location','SE');
xlabel('False positive rate');
ylabel('True positive rate');
xlim([-0.05 0.65]);
ylim([0 1.1]);
grid on;

%% Focus on low FPR and TPR region (score is above 0)
keep = SfitLDA>0;
SfitLDA_2 = SfitLDA(keep);
Y_2 = Y(keep);
isGood_2 = strcmp(Y_2,'g');
isBad_2 = strcmp(Y_2,'b');
Ngood_2 = sum(isGood_2);
Nbad_2 = sum(isBad_2);

keep = threLDA>0;
threLDA_2 = threLDA(keep);
fprLDA_2 = fprLDA(keep);
tprLDA_2 = tprLDA(keep);
%%
% Obtain quantiles for kernel bandwidth estimation
Qgood = quantile(SfitLDA_2(isGood_2),[0.25 0.75]);
Qbad = quantile(SfitLDA_2(isBad_2),[0.25 0.75]);
%%
% Estimate optimal kernel bandwidth
Hgood_2 = 1.58*(Qgood(2)-Qgood(1))/1.34/Ngood_2^(1/3)
Hbad_2 = 1.58*(Qbad(2)-Qbad(1))/1.34/Nbad_2^(1/3)
%%
% Compute smoothed TNR and FNR at 1000 points uniformly spaced from 0 to 1
thre = linspace(min(SfitLDA_2),max(SfitLDA_2),1000);
F_2 = ksdensity(SfitLDA_2(isBad_2),thre,'width',Hbad_2,'function','cdf');
G_2 = ksdensity(SfitLDA_2(isGood_2),thre,'width',Hgood_2,'function','cdf');
%%
% Plot empirical and smoothed FPR
figure;
plot(threLDA_2,fprLDA_2);
hold;
plot(thre,Nbad_2/Nbad*(1-F_2),'r--');
legend('Empirical','Smoothed','Location','SW');
xlabel('Threshold');
ylabel('False positive rate');
%%
% Plot empirical and smoothed TPR
figure;
plot(threLDA_2,tprLDA_2);
hold;
plot(thre,Ngood_2/Ngood*(1-G_2),'r--');
legend('Empirical','Smoothed','Location','SW');
xlabel('Threshold');
ylabel('True positive rate');
%%
% Plot empirical and smoothed ROC curve
figure;
plot(fprLDA_2,tprLDA_2);
hold;
plot(Nbad_2/Nbad*(1-F_2),Ngood_2/Ngood*(1-G_2),'r--');
legend('Empirical','Smoothed','Location','SE');
xlabel('False positive rate');
ylabel('True positive rate');

